#[cfg(test)]
mod tests;

use log::warn;
use serde_json;
use serde_json::json;
use std::collections::HashMap;
use std::error::Error;
use std::fmt;
use std::fs;
use std::path;

extern crate custom_error;
use custom_error::custom_error;

const JSON_POINTER_SEPARATOR: &str = "/";

custom_error! {DBError
    NotDirectory{path: String} = "Path to database should point at the directory: {path}",
    NoFile{group_name: String, file_name: String} = r#"There is no "{file_name}" file in group "{group_name}"."#,
    IncorrectPointer{pointer: String} = "Incorrect pointer is specified: {pointer}.",
}

enum ValueReference<'a> {
    Object(&'a mut serde_json::Map<String, serde_json::Value>, String),
    Array(&'a mut Vec<serde_json::Value>, usize),
}

type FileID<'a> = (&'a str, &'a str);

pub struct Group {
    name: String,
    files: HashMap<String, serde_json::Value>,
}

pub struct Database {
    storage_path: path::PathBuf,
    groups: Vec<Group>,
}

impl fmt::Display for Group {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        writeln!(f, "[{}]", self.name)?;
        for (name, contents) in self.files.iter() {
            writeln!(f, r#"  File "{}": {}"#, name, contents.to_string())?;
        }
        Ok(())
    }
}

impl fmt::Display for Database {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        writeln!(f, "DB: {}", self.storage_path.display())?;
        for g in self.groups.iter() {
            writeln!(f, "{}", g)?;
        }
        Ok(())
    }
}

impl Database {
    pub fn new(storage_path: &path::Path) -> Result<Database, Box<dyn Error>> {
        if !storage_path.is_dir() {
            return Err(Box::new(DBError::NotDirectory {
                path: storage_path.display().to_string(),
            }));
        }
        let mut groups = Vec::new();
        for entry in fs::read_dir(storage_path)? {
            let entry = entry?;
            let path = entry.path();
            if !path.is_dir() {
                warn!(
                    r#"File {} found where only group directories are supposed to be"#,
                    path.display()
                );
            } else {
                let group = load_group(&path)?;
                groups.push(group);
            }
        }
        Ok(Database {
            storage_path: storage_path.to_path_buf(),
            groups,
        })
    }

    pub fn group_names(&self) -> Vec<String> {
        self.groups.iter().map(|x| x.name.clone()).collect()
    }

    pub fn file_names_in(&self, group_name: &str) -> Vec<String> {
        match self.groups.iter().find(|x| x.name.eq(group_name)) {
            Some(group) => group.files.keys().map(|x| x.clone()).collect(),
            None => Vec::new(),
        }
    }

    fn group_index(&self, group_name: &str) -> Option<usize> {
        self.groups.iter().position(|x| x.name.eq(group_name))
    }

    fn as_json_mut(&mut self, (group_name, file_name): FileID) -> Option<&mut serde_json::Value> {
        match self.group_index(group_name) {
            Some(index) => self.groups[index].files.get_mut(&file_name.to_owned()),
            _ => None,
        }
    }

    pub fn as_json(&self, (group_name, file_name): FileID) -> Option<&serde_json::Value> {
        match self.group_index(group_name) {
            Some(index) => self.groups[index].files.get(&file_name.to_owned()),
            _ => None,
        }
    }

    pub fn as_string(&self, file_id: FileID) -> Option<String> {
        self.as_json(file_id).map(|x| x.to_string())
    }

    pub fn get_json(&self, file_id: FileID, pointer: &str) -> Option<&serde_json::Value> {
        match self.as_json(file_id) {
            Some(v) => v.pointer(pointer),
            _ => None,
        }
    }

    pub fn get_string(&self, file_id: FileID, pointer: &str) -> Option<String> {
        self.get_json(file_id, pointer).map(|x| x.to_string())
    }

    pub fn contains(&self, file_id: FileID, pointer: &str) -> bool {
        self.get_json(file_id, pointer) != None
    }

    pub fn set_json(
        &mut self,
        (group_name, file_name): FileID,
        pointer: &str,
        new_value: serde_json::Value,
    ) -> Result<(), Box<dyn Error>> {
        let file_json = match self.as_json_mut((group_name, file_name)) {
            Some(file_json) => file_json,
            _ => {
                return Err(Box::new(DBError::NoFile {
                    group_name: group_name.to_owned(),
                    file_name: file_name.to_owned(),
                }))
            }
        };
        touch(file_json, pointer)?;
        match file_json.pointer_mut(pointer) {
            Some(v) => *v = new_value,
            _ => {
                //  If after `touch()` call we still don't have an existing value -
                //  something is wrong with the `pointer`
                return Err(Box::new(DBError::IncorrectPointer {
                    pointer: pointer.to_owned(),
                }));
            }
        };
        Ok(())
    }

    pub fn remove(
        &mut self,
        (group_name, file_name): FileID,
        pointer: &str,
    ) -> Result<(), Box<dyn Error>> {
        let file_json = match self.as_json_mut((group_name, file_name)) {
            Some(file_json) => file_json,
            _ => {
                return Err(Box::new(DBError::NoFile {
                    group_name: group_name.to_owned(),
                    file_name: file_name.to_owned(),
                }))
            }
        };
        match pointer_to_reference(file_json, pointer) {
            Some(ValueReference::Object(map, variable_name)) => {
                map.remove(&variable_name);
            }
            Some(ValueReference::Array(vec, variable_index)) => {
                if variable_index < vec.len() {
                    vec.remove(variable_index);
                }
            }
            _ => {
                return Err(Box::new(DBError::IncorrectPointer {
                    pointer: pointer.to_owned(),
                }))
            }
        };
        Ok(())
    }
}

fn load_group(group_path: &path::Path) -> Result<Group, Box<dyn Error>> {
    let mut files = HashMap::new();
    for entry in fs::read_dir(group_path)? {
        let entry = entry?;
        let path = entry.path();
        if path.is_dir() {
            warn!(
                r#"Directory {} found where only data files are supposed to be"#,
                path.display()
            );
        } else {
            let file_contents = fs::read_to_string(&path)?;
            let file_name = get_file_name(path.as_path());
            let file_contents = serde_json::from_str(&file_contents)?;
            files.insert(file_name, file_contents);
        }
    }
    Ok(Group {
        name: get_file_name(group_path),
        files,
    })
}

fn touch(json_root: &mut serde_json::Value, pointer: &str) -> (Result<(), Box<dyn Error>>) {
    if pointer.is_empty() || json_root.pointer_mut(pointer).is_some() {
        return Ok(());
    }
    match pointer_to_reference(json_root, pointer) {
        Some(ValueReference::Object(map, variable_name)) => {
            map.insert(variable_name, json!(null));
        }
        Some(ValueReference::Array(vec, variable_index)) => {
            //  Since values at the index does not exist - resize will increase the sizeof the array
            vec.resize(variable_index + 1, json!(null));
        }
        _ => {
            return Err(Box::new(DBError::IncorrectPointer {
                pointer: pointer.to_owned(),
            }))
        }
    };
    Ok(())
}

fn pointer_to_reference<'a>(
    json_root: &'a mut serde_json::Value,
    pointer: &str,
) -> Option<ValueReference<'a>> {
    if pointer.is_empty() {
        return None;
    }
    let container_variable_pair =
        pop_json_pointer(pointer).and_then(move |(path, variable_name)| {
            match json_root.pointer_mut(&path) {
                Some(v) => Some((v, variable_name)),
                _ => None,
            }
        });
    let (json_container, variable_name) = match container_variable_pair {
        Some(v) => v,
        _ => return None,
    };
    match json_container {
        serde_json::Value::Object(map) => Some(ValueReference::Object(map, variable_name)),
        serde_json::Value::Array(vec) => {
            let index: usize = match variable_name.parse() {
                Ok(v) => v,
                _ => return None,
            };
            Some(ValueReference::Array(vec, index))
        }
        _ => None,
    }
}

fn pop_json_pointer(pointer: &str) -> Option<(String, String)> {
    let mut pointer = pointer.to_string();
    let last_separator_index = match pointer.rfind(JSON_POINTER_SEPARATOR) {
        Some(v) => v,
        _ => {
            return None;
        }
    };
    if last_separator_index >= pointer.len() {
        pointer.pop();
        return Some((pointer, String::new()));
    }
    let var_name = pointer.split_off(last_separator_index + 1);
    pointer.pop();
    Some((pointer, var_name))
}

fn get_file_name(path: &path::Path) -> String {
    path.file_stem()
        .and_then(|x| x.to_str())
        .unwrap_or_default()
        .to_string()
}
// TODO add tests for remove
// TODO add tests for panics (both add and remove)
// TODO add file addition/removal
// TODO add db saving

// TODO make sure file's main value not being an object won't break anything
// TODO check that file name is appropriate
// TODO handle parsing errors differently
// TODO add logs
